#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include "light.h"

void init_pointlight(struct light *lt, float x, float y, float z)
{
	lt->x = x;
	lt->y = y;
	lt->z = z;
	lt->tx = lt->ty = lt->tz = 0.0f;
	lt->r = lt->g = lt->b = 1.0f;
	lt->fov = 30.0;
}

void init_spotlight(struct light *lt, float x, float y, float z, float tx, float ty, float tz)
{
	lt->x = x;
	lt->y = y;
	lt->z = z;
	lt->tx = tx;
	lt->ty = ty;
	lt->tz = tz;
	lt->r = lt->g = lt->b = 1.0f;
	lt->fov = 30.0;
}

void set_pointlight(struct light *lt)
{
	float pos[4], col[4];

	pos[0] = lt->x;
	pos[1] = lt->y;
	pos[2] = lt->z;
	pos[3] = 1.0;

	col[0] = lt->r;
	col[1] = lt->g;
	col[2] = lt->b;
	col[3] = 1.0;

	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, col);
	glLightfv(GL_LIGHT0, GL_SPECULAR, col);
}

void set_spotlight(struct light *spot)
{
	float dir[4];

	set_pointlight(spot);

	dir[0] = spot->tx - spot->x;
	dir[1] = spot->ty - spot->y;
	dir[2] = spot->tz - spot->z;
	dir[3] = 1.0;

	glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, dir);
	glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, spot->fov);

	/* used as angle between inner/outer cone in the shader */
	glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, 5);
}
