#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include "light.h"
#include "sdr.h"

#define ROT_SCALE	0.4
#define PAN_SCALE	0.1
#define DOLLY_SCALE	0.1

#define DEPTH_SZ	512

void redraw(void);
void draw_geometry(void);
void draw_plane(float sz, int subdiv);
void draw_teapot(float sz);
void draw_sphere(float x, float y, float z, float rad);
void set_mat(float r, float g, float b, float spec, float shin);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mbutton(int bn, int state, int x, int y);
void mmotion(int x, int y);
void timer_func(int foo);

static int xsz, ysz;

static float cam_x, cam_y, cam_z;
static float cam_yaw = 60, cam_pitch = 30;
static float cam_dolly = 13;
static int show_depth = 0;

static struct light lt;
static unsigned int sdr;

static unsigned int fbo, dtex;
static unsigned int msec;

int main(int argc, char **argv)
{
	float amb[] = {0.05, 0.05, 0.05, 1.0};

	glutInitWindowSize(1024, 768);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("shadows test");

	glewInit();

	xsz = glutGet(GLUT_WINDOW_WIDTH);
	ysz = glutGet(GLUT_WINDOW_HEIGHT);

	glutDisplayFunc(redraw);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mbutton);
	glutMotionFunc(mmotion);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb);

	init_spotlight(&lt, 0, 10, 10, 0, 0, 0);
	lt.fov = 25.0;

	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0);

	glPolygonOffset(-1.0, 1.0);

	if(!(sdr = create_program_load("sdr.v.glsl", "sdr.p.glsl"))) {
		return 1;
	}

	/* create the render target depth texture */
	glGenTextures(1, &dtex);
	glBindTexture(GL_TEXTURE_2D, dtex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24, DEPTH_SZ, DEPTH_SZ, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, 0);

	/* create FBO */
	glGenFramebuffersEXT(1, &fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, dtex, 0);
	glDrawBuffer(GL_FALSE);
	glReadBuffer(GL_FALSE);

	if(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT) {
		fprintf(stderr, "incomplete framebuffer object\n");
		return 1;
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glutMainLoop();
	return 0;
}

void depth_pass(void)
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* setup the light projection matrix */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(lt.fov * 2.0, 1.0, 1.0, 100.0);

	/* ... and the viewport matrix for the rendertarget */
	glViewport(0, 0, DEPTH_SZ, DEPTH_SZ);

	/* setup the light view matrix */
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(lt.x, lt.y, lt.z, lt.tx, lt.ty, lt.tz, 0, 1, 0);

	glCullFace(GL_FRONT);

	/* render to texture */
	draw_geometry();

	glCullFace(GL_BACK);

	glViewport(0, 0, xsz, ysz);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void redraw(void)
{
	msec = glutGet(GLUT_ELAPSED_TIME);

	glutTimerFunc(16, timer_func, 0);

	/* render depth buffer */
	depth_pass();

	/* ---- render scene ---- */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	gluPerspective(lt.fov * 2.0, 1.0, 1.0, 100.0);
	gluLookAt(lt.x, lt.y, lt.z, lt.tx, lt.ty, lt.tz, 0, 1, 0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -cam_dolly);
	/*glTranslatef(cam_x, cam_y, cam_z);*/
	glRotatef(cam_pitch, 1, 0, 0);
	glRotatef(cam_yaw, 0, 1, 0);

	set_spotlight(&lt);

	/*glPolygonOffset(2, 2);
	glEnable(GL_POLYGON_OFFSET_FILL);*/

	bind_program(sdr);
	draw_geometry();
	bind_program(0);

	/*glDisable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0, 0);*/

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);

	if(show_depth) {
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glBindTexture(GL_TEXTURE_2D, dtex);
		glEnable(GL_TEXTURE_2D);

		glBegin(GL_QUADS);

		glTexCoord2f(0, 0);
		glVertex2f(-0.9, -0.9);
		glTexCoord2f(1, 0);
		glVertex2f(-0.2, -0.9);
		glTexCoord2f(1, 1);
		glVertex2f(-0.2, -0.2);
		glTexCoord2f(0, 1);
		glVertex2f(-0.9, -0.2);

		glEnd();

		glDisable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, 0);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_LIGHTING);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

	glutSwapBuffers();
}

void draw_geometry(void)
{
	set_mat(0.2, 0.7, 0.3, 0.4, 30.0);
	draw_plane(20.0, 10);

	set_mat(0.2, 0.4, 0.8, 1.0, 70.0);
	draw_teapot(1.5);

	set_mat(1.0, 0.4, 0.3, 0.9, 50.0);
	draw_sphere(-2, 1.0, -3, 1.0);
}

void draw_plane(float sz, int subdiv)
{
	int i, j;
	float sep = sz / (float)(subdiv + 1);

	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);

	for(i=0; i<subdiv + 1; i++) {
		for(j=0; j<subdiv + 1; j++) {
			float x = (float)i * sep - sz * 0.5;
			float y = (float)j * sep - sz * 0.5;

			glVertex3f(x, 0, y);
			glVertex3f(x, 0, y + sep);
			glVertex3f(x + sep, 0, y + sep);
			glVertex3f(x + sep, 0, y);
		}
	}

	glEnd();
}

void draw_teapot(float sz)
{
	float t = msec / 10.0;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(0, sz + 0.3 * sz, 0);
	glRotatef(t * 0.5, 0, 1, 0);

	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glTranslatef(0, sz + 0.3 * sz, 0);
	glRotatef(t * 0.5, 0, 1, 0);

	glutSolidTorus(0.3 * sz, sz, 10, 30);
	/*
	glBegin(GL_TRIANGLES);
	glNormal3f(0, 0.5, 1);
	glVertex3f(0.7, -0.75 * sz, 0.7);
	glVertex3f(0, 0.7, 0);
	glVertex3f(-0.7, -0.75 * sz, 0.7);

	glNormal3f(0, 0.5, -1);
	glVertex3f(-0.7, -0.75 * sz, -0.7);
	glVertex3f(0, 0.7, 0);
	glVertex3f(0.7, -0.75 * sz, -0.7);

	glNormal3f(1, 0.5, 0);
	glVertex3f(-0.7, -0.75 * sz, 0.7);
	glVertex3f(0, 0.7, 0);
	glVertex3f(-0.7, -0.75 * sz, -0.7);

	glNormal3f(-1, 0.5, 0);
	glVertex3f(0.7, -0.75 * sz, -0.7);
	glVertex3f(0, 0.7, 0);
	glVertex3f(0.7, -0.75 * sz, 0.7);
	glEnd();
	*/

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void draw_sphere(float x, float y, float z, float rad)
{
	float t = msec / 1000.0;

	x = cos(t) * 3.0;
	z = sin(t) * 3.0;
	y += fabs(sin(t * 4.0) * 2.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glTranslatef(x, y, z);

	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glTranslatef(x, y, z);

	glutSolidSphere(rad, 20, 10);

	glMatrixMode(GL_TEXTURE);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void set_mat(float r, float g, float b, float spec, float shin)
{
	float col[4], scol[4];

	col[0] = r;
	col[1] = g;
	col[2] = b;
	col[3] = 1.0;

	scol[0] = scol[1] = scol[2] = spec;
	scol[3] = 1.0;

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, col);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, scol);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shin);
}

void reshape(int x, int y)
{
	printf("reshape called\n");
	glViewport(0, 0, x, y);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 1.0, 1000.0);
	glMatrixMode(GL_MODELVIEW);

	xsz = x;
	ysz = y;
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 'd':
		show_depth = !show_depth;
		glutPostRedisplay();
		break;

	case 27:
	case 'q':
	case 'Q':
		exit(0);

	default:
		break;
	}
}

static int mbtn;
static int prev_x, prev_y;
void mbutton(int bn, int state, int x, int y)
{
	prev_x = state == GLUT_DOWN ? x : -1;
	prev_y = state == GLUT_DOWN ? y : -1;
	mbtn = state == GLUT_DOWN ? bn : -1;
}

void mmotion(int x, int y)
{
	int dx, dy;

	dx = x - prev_x;
	dy = y - prev_y;

	switch(mbtn) {
	case GLUT_LEFT_BUTTON:
		/* rotate */
		cam_yaw += (float)dx * ROT_SCALE;
		cam_pitch += (float)dy * ROT_SCALE;

		if(cam_pitch > 90.0) cam_pitch = 90.0;
		if(cam_pitch < -90.0) cam_pitch = -90.0;
		break;

	case GLUT_MIDDLE_BUTTON:
		/* pan */
		cam_x += (float)dx * PAN_SCALE;
		cam_z += (float)dy * PAN_SCALE;
		break;

	case GLUT_RIGHT_BUTTON:
		/* zoom */
		cam_dolly += (float)dy * DOLLY_SCALE;
		if(cam_dolly < 0.0) cam_dolly = 0.0;
		break;

	default:
		return;	/* to avoid updating prev_x/prev_y and redrawing */
	}

	prev_x = x;
	prev_y = y;
	glutPostRedisplay();
}

void timer_func(int foo)
{
	glutPostRedisplay();
}
